﻿# broadway_base.py
#
# This file is part of python wrapper for the BROADWAY2 package.
#
# MIT License

# Copyright (c) 2013-2022 FibreCode GmbH <info@fibrecode.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import platform
import datetime
import time
import threading
import ctypes

from broadway2 import broadway_api, fca

BROADWAY_VERSION_MAJOR = 2
BROADWAY_VERSION_MINOR = 20 # 2022-01-28 Requires broadway2-FW >= V2.30

TIME_ZERO = datetime.datetime(1970, 1, 1)

def Timestamp64():
    """
    return UTC-timstamp in ms since 1970-1-1
    """
    return int((datetime.datetime.utcnow() - TIME_ZERO).total_seconds() * 1000)


def factor_to_int32(factor):
    """fixed-point decimal with 32-bit: 1.xxxxxxxx (max. 1.9999999) 
    hexVal = 0x80000000 : 1.0000
    if factor is out of valid range, default 0x80000000 returned
    """
    if factor < 0 or factor > 2:
        return 0x80000000

    return int(factor*2**31)


def int32_to_factor(intValue):
    """
    intValue 0x80000000 returns 1.0
    """
    return intValue/2**31

class HWTimestamp:
    """HW Timestamp inside USB-Raw Adapters like FC611, FC612
    uint64: seconds
    int32: nano-seconds
    """
    def __init__(self, secs=0, nanosecs=0):
        self.seconds = int(secs)
        self.nanosec = int(nanosecs)

    def from_tuple(self, tmTuple):
        """tmTuple must be tuple of 2
        [0] seconds
        [1] Nanoseconds
        """
        try:
            self.seconds = tmTuple[0]
        except:
            self.seconds = 0
        try:
            self.nanosec = tmTuple[1]
        except:
            self.nanosec = 0

        return self


    def __str__(self):
        return '{0}.{1:09}'.format(self.seconds, self.nanosec)

    def get_timestamp_ns(self):
        return self.seconds *1e9 + self.nanosec

    def __add__(self, hwTm):
        """overload add operator"""
        sec = self.seconds + hwTm.seconds
        nano = self.nanosec + hwTm.nanosec
        if nano > 1e9:
            sec += 1
            nano -= 1e9

        return HWTimestamp(sec, nano)

    def __sub__(self, other):
        """overload substract operator"""
        tm1 = 1e9*other.seconds + other.nanosec
        tm2 = 1e9*self.seconds + self.nanosec

        diff = tm2 - tm1
        if diff < 0:
            sec = diff // 1e9 + 1
            nano = sec*1e9  - diff
        else:
            sec = diff // 1e9
            nano = diff - sec*1e9 

        return HWTimestamp(sec, nano)

    def __lt__(self, other):
        """overload less-than operator"""
        tm1 = 1e9*other.seconds + other.nanosec
        tm2 = 1e9*self.seconds + self.nanosec

        diff = tm2 - tm1
        if diff < 0:
            return True
        else:
            return False

    def __le__(self, other):
        """overload less-than equal operator"""
        tm1 = 1e9*other.seconds + other.nanosec
        tm2 = 1e9*self.seconds + self.nanosec

        diff = tm2 - tm1
        if diff <= 0:
            return True
        else:
            return False

    def __gt__(self, other):
        """overload less-than operator"""
        tm1 = 1e9*other.seconds + other.nanosec
        tm2 = 1e9*self.seconds + self.nanosec

        diff = tm2 - tm1
        if diff > 0:
            return True
        else:
            return False

    def __eq__(self, other):
        """overload equal operator"""

        if other.seconds == self.seconds and other.nanosec == self.nanosec:
            return True
        else:
            return False

    def __ne__(self, other):
        """overload not equal operator"""

        if other.seconds != self.seconds or other.nanosec != self.nanosec:
            return True
        else:
            return False

class RawRxHeaderTypes:
    HeaderType_Plain = 0
    #HeaderType1 (pre-dev, no more supported)
    HeaderType2 = broadway_api.RawRxHeaderType.Type2    # adds 64+32-bit Timestamp and 32 bit SeqCounter

class MetaSingleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(MetaSingleton, cls).__call__(*args, **kwargs)
            return cls._instances[cls]

class SingletonMixin(object):
    """A singleton mixin which can be subclassed to create singleton classes"""
    __metaclass__ = MetaSingleton


class Broadway(SingletonMixin):
    """Singelton as basic instance for all adapters"""
    def __init__(self, libPath=''):
        self.__libPath = libPath
        self.__adapters = []    # list of current connected adapters
        self.__oldAdapters = []

        if not self.__initialize():
            raise FileNotFoundError('broadway2-API not found')
        self.__update_adapter_info()

    def __initialize(self):
        """initiaize dll setup and verify api-version"""
        res = broadway_api.initialize(self.__libPath)
        if res == False:
            return False

        ver = broadway_api.get_api_version()
        self.__api_version = ver

        res = broadway_api.check_api_version(BROADWAY_VERSION_MAJOR, BROADWAY_VERSION_MINOR)
        
        if res == False:
            return False

        self.extNotifyFunc = None
        self.cbNotify = broadway_api.CB_DeviceChangeEvent(self.notify_device_change)
        self.__notifyHandle = broadway_api.register_device_change(self.cbNotify)

        self.initialized = True

        return True

    def get_unsupported_adapters(self):
        return self.__oldAdapters

    def set_notify(self, notifyFunc):
        self.extNotifyFunc = notifyFunc
        # call direct back to be up-to-date
        self.extNotifyFunc(0)

    def clear_notify(self):
        self.extNotifyFunc = None
        

    def notify_device_change(self, context, event):
        self.scan_and_update()
        if self.extNotifyFunc != None:
            self.extNotifyFunc(event)

    def __update_adapter_info(self):
        """scan all connected devices and create/update fcb list"""
        self.__adapters = []
        self.__oldAdapters = []

        for inst in range(0, broadway_api.get_num_devices()):
            dev = broadway_api.open_control(broadway_api.FCBDeviceType.FCBDeviceTypeAny, inst)
            if dev != None:
                # select correct FCAdapter
                
                a = fca.FCAdapter()
                a.initialize(inst, dev)   # update all FCAdapter instance properties to sync with connected USB Adapter
                
                self.__adapters.append(a)

                broadway_api.close(dev) # device is closed
            else:
                if broadway_api.LastResult == -1:
                    self.__oldAdapters.append(inst)
                

    def scan_and_update(self):
        """re-initialize and scan again connected adapters
        First scan is automatically done on __init__
        """
        if self.initialized:
            res = True
        else:
            res = self.__initialize()
        if res:
            self.__update_adapter_info()
            return True

        return False

    @property
    def Count(self):
        """return number of found adapters"""
        return len(self.__adapters)

    def enumerate(self):
        """return generator to walk through all connected adapters"""
        for a in self.__adapters:
            yield a

    def get_adapter(self, instance):
        """return adapter given by index number in list"""
        res = None
        for a in self.__adapters:
            if a.Inst == instance:
                res = a
                break
        return res

    def get_adapters(self):
        """return list of Adapters as tuple (inst, friendlyName)"""
        res = []
        for a in self.__adapters:
            res.append((a.Inst, a.FriendlyName))
        return res

    def get_raw_adapters(self):
        """return list of tuple of Raw-Adapters"""
        res = []
        for a in self.__adapters:
            if a.Features['Raw']:
                res.append((a.Inst, a.FriendlyName))
        return res

    def get_tool_adapters(self):
        """return list of tuple of Tool-Adapters (support for SPI/MDIO)"""
        res = []
        for a in self.__adapters:
            if a.Features['Tool']:
                res.append((a.Inst, a.FriendlyName))
        return res

    def get_adapters_by_type(self, deviceType):
        """return list of Adapters matching given type"""
        res = []
        for a in self.__adapters:
            if a.DeviceType == deviceType:
                res.append((a.Inst, a.FriendlyName))
        return res

    def get_adapter_by_name(self, friendlyName):
        """return adapter instance if found by given friendlyName, else returns None"""
        for a in self.__adapters:
            if friendlyName == a.FriendlyName:
                return a
        return None

    def open_control(self, instId):
        """open adapter instance, and return:
        in case of Error: None
        in case of Success: dev
        """
        res = broadway_api.open_control(broadway_api.FCBDeviceType.FCBDeviceTypeAny, instId)
        
        return res

    def open_raw_tx(self, instId):
        """open adapter instance, and return:
        in case of Error: None
        in case of Success: dev
        """
        res = broadway_api.open_raw_tx(broadway_api.FCBDeviceType.FCBDeviceTypeAny, instId)
        
        return res

    def open_raw_rx(self, instId, headerType=0, buffers=0):
        """open adapter instance, and return:
        in case of Error: None
        in case of Success: dev
        """
        res = broadway_api.open_raw_rx(broadway_api.FCBDeviceType.FCBDeviceTypeAny, instId, headerType, buffers)
        
        return res

    def close(self, dev):
        """try to close connected adapter instance"""
       
        res = broadway_api.close(dev)
        
        return True

    def get_inst_from_board(self, boardName):
        """find connected adapter in 2 options
        FCxxxx-xx-nnnnnn
        FCxxxx-xx
        Second option only returns successfully, if there is only one board of this type connected"""
        parts = boardName.split('-')
        if len(parts) == 3:
            for a in self.__adapters:
                if (a.SerialNumber == int(parts[2])) and (a.PcbType == parts[0]) and (a.PcbOpt == parts[1]):
                    return a.Inst
        elif len(parts) == 2:
            res = []
            for fcinfo in self.__fcInfo:
                if (fcinfo.PcbType == parts[0]) and (fcinfo.PcbOpt == parts[1]):
                    res.append(fcinfo.Inst)
            if len(res) == 1:
                return res[0]
            
        return None
